!
! Module: esrcselect_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Select EPIC sources for source product generation
! 
! TODO:
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esrcselect_mod

  use dal
  use param  
  use errorHandling
  use datafilter
  use exprcolgen	! in selectlib

  use dal_aux		! for splitSetTabName, chooseRandomNewColName
  use sort_mod
  use ssc_misc		! for intToStr

  implicit none 
  
  private 

  ! Command-line parameters:
  character(256) ::&
    tempSetName = '',&
    inSrcListSetTabName = '',&
    outSrcListSetName = '',&
    selectionExpression = '',&
    maxSourceNumSortExpression = ''
  character(20) ::&
!    selectStyle = '',&
    userIdColName = '',&
    userFlagColName=''
  integer ::&
    maxSourceNum = 0,&
    userIdInst=0,&
    userIdBand=0,&
    userFlagBit=0
  logical ::&
    withIdColName = .false.,&
    withSort=.false.,&
    deleteUnselected=.false.

  public esrcselect

contains 
  
  !---------------------------------------------------------------------
  subroutine esrcselect
  !---------------------------------------------------------------------

    implicit none

    character(256) :: inSrcListSetName, preFilterExpression
    character(8) :: srcListTabName, maxSrcNumStr
    character(2) :: idBandStr, idInstStr
    type(DataSetT) :: sourcesSet, outSet
    type(TableT) :: sourcesTab, outTab
    integer :: numSourcesIn, numFilteredRows, i
    character(20) :: sortColName, indexColName, flagColName
    type(ColumnT) :: indexCol, sortCol
    integer(int32), pointer :: indexData(:)
    real(double), pointer :: sortColData(:)
    integer(int32), allocatable :: tempIndices(:)
    integer(int32) :: flagMaskInteger

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call splitSetTabName(inSrcListSetTabName, inSrcListSetName, srcListTabName)
    ! in ssclib/dal_aux.

    ! Have to filter the source list, then optionally sort it and choose the
    ! first N. Both processes involve an expression which relates various
    ! column names in the input source list: in the filter case, this
    ! expression must evaluate to a logical value; in the sort case, to
    ! a numeric quantity. However, the trick to esrcselect is that _only_
    ! values from rows of the specified values of ID_INST and ID_BAND
    ! are included in the evaluation of the column expression.
    !
    ! The first step is thus to pre-filter the input list, removing all
    ! rows except those with the right combination of ID_INST and ID_BAND:

    idBandStr = intToStr(userIdBand) ! in ssclib/ssc_misc.
    idInstStr = intToStr(userIdInst) ! in ssclib/ssc_misc.
    preFilterExpression = '(ID_INST=='//trim(idInstStr)&
      //')&&(ID_BAND=='//trim(idBandStr)//')'

    sourcesSet = clone(inSrcListSetName, tempSetName, TEMP)
    sourcesTab = table(sourcesSet, srcListTabName, MODIFY)
    numSourcesIn = numberOfRows(sourcesTab)
    if (numSourcesIn <= 0)&
      call warning('noSources',"The input list contained &
      &no sources.")

    call initFilter(preFilterExpression)
    numFilteredRows = doFilter(sourcesTab)

    if (numFilteredRows <= 0)&
      call warning('noQualifyingSources'&
      , "The input list contained no sources which have the specified &
      &combination of --idinst and --idband.")

!*** check each value of *_ID_SRC??

    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourcesTab)

    if (numFilteredRows <= 0) then
      call warning('noSourcesSelected'&
        , "No sources passed the specified filtering expression.")

    else if (numFilteredRows == 1) then
      call message(AppMsg, Verbose, '1 source passed the &
        &selection expression.')

    else
      write (errStr, '(i6, a)') numFilteredRows, ' sources passed the &
        &selection expression.'
      call message(AppMsg, Verbose)
    end if

    if (withSort) then
      sortColName = chooseRandomNewColName(sourcesTab) ! in ssclib/dal_aux.
      call initExprColGen(maxSourceNumSortExpression, sortColName, REAL64)
      call doExprColGen(sourcesTab, sourcesTab)

      indexColName = chooseRandomNewColName(sourcesTab) ! in ssclib/dal_aux.
      indexCol = addColumn(sourcesTab, indexColName, INTEGER32)
      indexData => int32Data(indexCol)

      sortCol = column(sourcesTab, sortColName)
      sortColData => real64Data(sortCol)

      allocate(tempIndices(numFilteredRows))
      tempIndices = (/(i, i=1,numFilteredRows)/)

      call quickSort(sortColData, tempIndices) ! in ssclib/sort_mod.
      do i = 1, numFilteredRows
        indexData(tempIndices(i) - 1) = i
      end do
      call release(sortCol)
      call release(indexCol)

      maxSrcNumStr = intToStr(maxSourceNum) ! in ssclib/ssc_misc.
      selectionExpression = trim(indexColName)//"<="//trim(maxSrcNumStr)
      call initFilter(selectionExpression)
      numFilteredRows = doFilter(sourcesTab)
if (numFilteredRows <= 0) call error('bug0',"Please contact the code developer.")

    end if ! withSort

    ! sourcesTab now contains just those source IDs which we want to
    ! retain in the output. The simplest way to filter the output is
    ! to create a flag column in the output, set those rows of it for
    ! which the source ID is found in sourcesTab, then delete all unset
    ! rows.

    if (inSrcListSetName .eq. outSrcListSetName) then
      outSet = dataSet(inSrcListSetName, MODIFY)
    else
      outSet = clone(inSrcListSetName, outSrcListSetName, MODIFY)
    end if
    outTab = table(outSet, srcListTabName, MODIFY)

    call saveFlagData(outTab, sourcesTab, flagColName, flagMaskInteger)

    write (selectionExpression, '(a, i7, a)')&
      '('//trim(flagColName)//' & ', flagMaskInteger, ')!=0'
    call initFilter(selectionExpression)

    if (deleteUnselected) then
      numFilteredRows = doFilter(outTab)
!if (numFilteredRows <= 0) call error('bug1',"Please contact the code developer.")

      call deleteColumn(outTab, flagColName)
    end if

    call release(outTab)
    call release(sourcesTab)
    call release(outSet)
    call release(sourcesSet)

  end subroutine esrcselect
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    tempSetName = stringParameter('tempset')
    inSrcListSetTabName = stringParameter('table')
    outSrcListSetName = stringParameter('filteredset')
    selectionExpression = stringParameter('expression')

    withSort = booleanParameter('withsort')
    if (withSort) then
      maxSourceNumSortExpression = stringParameter('sortexpression')
      maxSourceNum = intParameter('maxn')
    end if

    deleteUnselected = booleanParameter('destruct')
    if (.not.deleteUnselected) then
      userFlagColName = stringParameter('flagcolumn')
      userFlagBit = intParameter('flagbit')
    end if

    withIdColName = booleanParameter('withidcol')
    if (withIdColName)&
      userIdColName = stringParameter('idcol')

    userIdInst = intParameter('idinst')
    userIdBand = intParameter('idband')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine saveFlagData(outTab, sourcesTab, flagColName, flagMaskInteger)
  !---------------------------------------------------------------------

!**** this is similar to a subroutine in confusionselect.

    implicit none

    type(TableT),   intent(in)  :: outTab,&
                                   sourcesTab
    character(*),   intent(out) :: flagColName
    integer(int32), intent(out) :: flagMaskInteger

    logical :: initializeFlagsToZero
    character(20) :: idColName
    integer :: flagBit, flagDataType, row
    type(ColumnT) :: flagColumn
    integer(int32), pointer :: flagDataInt32(:), tempIdData(:), outIdData(:)
    integer(int16), pointer :: flagDataInt16(:)
    integer(int8), pointer :: flagDataInt8(:)

    !-------------------------------------------------------------------
    ! Start

    initializeFlagsToZero = .true. ! default

    if (deleteUnselected) then ! our flag column will also be deleted.
    ! Just choose any old name and data type.

      flagColName = chooseRandomNewColName(outTab)
      flagBit = 0
      flagDataType = INTEGER8
      flagColumn = addColumn(outTab, flagColName, flagDataType)

    else
      flagColName = userFlagColName
      flagBit     = userFlagBit

      if (hasColumn(outTab, userFlagColName)) then
        flagColumn = column(outTab, userFlagColName)
        flagDataType = dataType(flagColumn)

        if (  flagDataType .ne. INTEGER8&
        .and. flagDataType .ne. INTEGER16&
        .and. flagDataType .ne. INTEGER32)&
call error('',"")

        initializeFlagsToZero = .false. ! don't want to overwrite info
        ! that is already there.

      else
        flagDataType = INTEGER8
!*** allow user to specify data type (then also require existing datatype to equal this if column already exists?)?
        flagColumn = addColumn(outTab, flagColName, flagDataType)
      end if
    end if

    if (withIdColName) then
      idColName = userIdColName
    else
      if (hasColumn(sourcesTab, 'ML_ID_SRC')) then
        idColName = 'ML_ID_SRC'
      else
!*** warning
        idColName = 'BOX_ID_SRC'
      end if
    end if
    tempIdData => int32Data(column(sourcesTab, idColName))
    outIdData  => int32Data(column(outTab,     idColName))

    select case(flagDataType)
    case(INTEGER8)
if (flagBit > 7) call error('',"")
      flagDataInt8 => int8Data(flagColumn)

      if (initializeFlagsToZero) then
        flagDataInt8 = 0
      else
        flagDataInt8 = ibclr(flagDataInt8, flagBit)
      end if

      do row = 0, size(tempIdData)-1
        where(outIdData == tempIdData(row))&
          flagDataInt8 = ibset(flagDataInt8, flagBit)
      end do

    case(INTEGER16)
if (flagBit > 15) call error('',"")
      flagDataInt16 => int16Data(flagColumn)

      if (initializeFlagsToZero) then
        flagDataInt16 = 0
      else
        flagDataInt16 = ibclr(flagDataInt16, flagBit)
      end if

      do row = 0, size(tempIdData)-1
        where(outIdData == tempIdData(row))&
          flagDataInt16 = ibset(flagDataInt16, flagBit)
      end do

    case(INTEGER32)
if (flagBit > 31) call error('',"")
      flagDataInt32 => int32Data(flagColumn)

      if (initializeFlagsToZero) then
        flagDataInt32 = 0
      else
        flagDataInt32 = ibclr(flagDataInt32, flagBit)
      end if

      do row = 0, size(tempIdData)-1
        where(outIdData == tempIdData(row))&
          flagDataInt32 = ibset(flagDataInt32, flagBit)
      end do

    case default
call error('',"")
    end select

    flagMaskInteger = 0
    flagMaskInteger = ibset(flagMaskInteger, flagBit)

    call release(flagColumn)

  end subroutine saveFlagData

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module esrcselect_mod
